#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#include "klient.h"

const int OK = 0;
const int ZLE_HASLO = 1;
const int SERWER_PELNY = 2;
const int BLAD = 3;
const int ANULOWANIE = 4;
const int WYSYLANIE = 5;
const int UAKTUALNIENIE = 6;

int main(int argc, char**argv) {
  int port=1023;
  int pd, sd, kd;
  int pos, i, j, k, msg;
  unsigned long len, dlugosc, granice[2];
  long roznica;
  char nazwa[128];
  char haslo_buf[17];
  char*txt, *pocztxt, *kontxt;
  struct sockaddr_in moj_adr;
  fd_set glowny;
  fd_set sel;
  struct klient klienci[50];
  
  if (argc<3) {
    printf("Parametry: plik port haslo\n");
    return EXIT_FAILURE;
  }

  sscanf(argv[2], "%d", &port);
  if (port<1024 || port>65535) {
    fprintf(stderr, "Nr portu musi byc z przedzialu 1024-65535\n");
    return EXIT_FAILURE;
  }

  if (strlen(argv[3])>16) {
    fprintf(stderr, "Maksymalna dlugosc hasla to 16 znakow\n");
    return EXIT_FAILURE;
  }

  if ((pd = open(argv[1], O_RDWR)) == -1) {
    perror(0);
    return EXIT_FAILURE;
  }

  FD_ZERO(&glowny);
  
  memset(klienci, 0, 50*sizeof(struct klient));
  
  memset(&moj_adr, 0, sizeof(struct sockaddr_in));	
  moj_adr.sin_family = AF_INET;
  moj_adr.sin_port = htons(port);
  moj_adr.sin_addr.s_addr = htonl(INADDR_ANY);

  sd = socket(PF_INET, SOCK_STREAM, 0);

  if (bind(sd, (struct sockaddr *)&moj_adr, sizeof(struct sockaddr))) {
    perror(0);
    close(sd);
    return EXIT_FAILURE;
  }

  listen(sd, 20);

  FD_SET(sd, &glowny);
  for (;;) {
    sel = glowny;
    select(FD_SETSIZE, &sel, 0, 0, 0);
    if (FD_ISSET(sd, &sel)) { // nowe polaczenie
      printf("Nowe polaczenie:\n");
      kd = accept(sd, 0, 0);

      pos = 0;
      memset(haslo_buf, 0, 17);
      do
        pos += read(kd, haslo_buf+pos, 1);
      while (haslo_buf[pos-1]!='\0' && pos<16);

      if (strcmp(argv[3], haslo_buf)) {
        writeall(kd, &ZLE_HASLO, sizeof(int));
        close(kd);
	printf("  zle haslo\n");
        continue;
      }

      for (i=0; i<50 && klienci[i].d; i++);
      if (i==50) {
        writeall(kd, &SERWER_PELNY, sizeof(int));
        close(kd);
	printf("  serwer pelny\n");
        continue;
      }
      
      klienci[i].d = kd;
      klienci[i].edytuje = 0;
      
      writeall(kd, &OK, sizeof(int));
      printf("  polaczono (%d)\n", kd);
      
      FD_SET(kd, &glowny);
      sendTxt(kd, pd, klienci);
    }
    
    fflush(stdout);
    for (i=0; i<50; i++)
      if (klienci[i].d && FD_ISSET(klienci[i].d, &sel)) {
      
      
      
      	if (!read(klienci[i].d, &msg, sizeof(int))) { // utrata polaczenia
	  close(klienci[i].d);
	  FD_CLR(klienci[i].d, &glowny);
	  memset(&klienci[i], 0, sizeof(struct klient));
	  for (j=0; j<50; j++)
	    if (klienci[j].d) {
	      write(klienci[j].d, &OK, sizeof(int));
	      sendTxt(klienci[j].d, pd, klienci);
	    }
	    
	    
	    
	} else if (!klienci[i].edytuje) { // klient zglasza fragment do edycji
	
	  read(klienci[i].d, granice, 2*sizeof(unsigned long));
	  
	  for (j=0; j<50; j++)
	    if (klienci[j].d && klienci[j].edytuje && i!=j &&
	      ((granice[0]>=klienci[j].pocz && granice[0]<=klienci[j].kon) ||
	       (granice[1]>=klienci[j].pocz && granice[1]<=klienci[j].kon)))
	      break;
	  
	  if (j==50) {
	    klienci[i].pocz = granice[0];
	    klienci[i].kon = granice[1];
	    klienci[i].edytuje = 1;
	    
	    for (j=0; j<50; j++)
	      if (klienci[j].d) {
	        write(klienci[j].d, &OK, sizeof(int));
	        sendTxt(klienci[j].d, pd, klienci);
              } 
	  } else
	    write(klienci[i].d, &BLAD, sizeof(int));
	    
	    
	    
	} else if (msg == ANULOWANIE) { // klient anuluje edycje
	  klienci[i].edytuje=0;
	  for (j=0; j<50; j++)
	    if (klienci[j].d) {
	      write(klienci[j].d, &OK, sizeof(int));
	      sendTxt(klienci[j].d, pd, klienci);
	    }
	    
	    
	    
	} else if (msg == WYSYLANIE) { // klient nadsyla zmodyfikowany fragment
	  dlugosc = getflen(pd);
	  read(klienci[i].d, &len, sizeof(unsigned long));
	  
	  txt = 0;
	  if (len) {
	    lseek(pd, klienci[i].pocz, SEEK_SET);
	    txt = (char*)malloc(len);
	    pos = 0;
	    while(pos!=len) {
      	      pos += read(klienci[i].d, &(txt[pos]), len-pos);
    	    }
	  }
	  
	  pocztxt = 0;
	  if (klienci[i].pocz) {
	    lseek(pd, 0, SEEK_SET);
	    pocztxt = (char*)malloc(klienci[i].pocz);
	    pos = 0;
	    while(pos!=klienci[i].pocz)
      	      pos += read(pd, &(pocztxt[pos]), klienci[i].pocz-pos);
	  }
	  
	  kontxt = 0;
	  if (klienci[i].kon<dlugosc-1) {
	    lseek(pd, klienci[i].kon+1, SEEK_SET);
	    kontxt = (char*)malloc(dlugosc-klienci[i].kon-1);
	    pos = 0;
	    while(pos!=dlugosc-klienci[i].kon-1)
      	      pos += read(pd, &(kontxt[pos]), dlugosc-klienci[i].kon-1-pos);
	  }
	  
	  close(pd);
	  pd = open(argv[1], O_RDWR | O_TRUNC);
	  
	  if (pocztxt) {
	    write(pd, pocztxt, klienci[i].pocz);
	    free(pocztxt);
	  }
	  if (txt) {
	    write(pd, txt, len);
	    free(txt);
	  }
	  if (kontxt) {
	    write(pd, kontxt, dlugosc-klienci[i].kon-1);
	    free(kontxt);
	  }
	  
	  printf("Plik zmodyfikowany\n");
	  fflush(stdout);
	  	  
	  roznica = len - (klienci[i].kon - klienci[i].pocz) - 1;
	  klienci[i].kon = klienci[i].pocz + len - 1;
	  
	  for (j=0; j<50; j++)
	    if (klienci[j].pocz>klienci[i].pocz) {
	      klienci[j].pocz += roznica;
	      klienci[j].kon += roznica;
	    }
	    
	  printf("Wysylanie uaktualnien:\n");
	  fflush(stdout);
	  
	  for (j=0; j<50; j++)
	    if (klienci[j].d) {
	      if (klienci[j].pocz <= klienci[i].pocz) {
	        write(klienci[j].d, &OK, sizeof(int));
	        sendTxt(klienci[j].d, pd, klienci);
	      } else {
	        write(klienci[j].d, &UAKTUALNIENIE, sizeof(int));
		write(klienci[j].d, &roznica, sizeof(long));
		sendTxt(klienci[j].d, pd, klienci);
	      }
            } 
	}
      }
  }
}
